import logging
import time
import uuid
import warnings

import docker
import psycopg2
import pytest
from dynamicannotationdb.annotation_client import DynamicAnnotationClient
from dynamicannotationdb.interface import DynamicAnnotationInterface
from dynamicannotationdb.materialization_client import DynamicMaterializationClient

logging.basicConfig(level=logging.DEBUG)
test_logger = logging.getLogger()


def pytest_addoption(parser):
    parser.addoption(
        "--docker",
        action="store",
        default=False,
        help="Use docker for postgres testing",
    )


@pytest.fixture(scope="session")
def docker_mode(request):
    return request.config.getoption("--docker")


def pytest_configure(config):
    config.addinivalue_line("markers", "docker: use postgres in docker")


@pytest.fixture(scope="session")
def database_metadata() -> dict:
    yield {
        "postgis_docker_image": "postgis/postgis:9.6-2.5",
        "db_host": "localhost",
        "sql_uri": f"postgresql://postgres:postgres@localhost:5432/test_volume",
    }


@pytest.fixture(scope="session")
def annotation_metadata():
    yield {
        "aligned_volume": "test_volume",
        "table_name": "anno_test",
        "schema_type": "synapse",
        "pcg_table_name": "test_pcg",
    }


@pytest.fixture(scope="session", autouse=True)
def postgis_server(docker_mode, database_metadata: dict) -> None:

    postgis_docker_image = database_metadata["postgis_docker_image"]
    sql_uri = database_metadata["sql_uri"]

    if docker_mode:
        test_logger.info(f"PULLING {postgis_docker_image} IMAGE")
        try:
            docker_client = docker.from_env()
            docker_client.images.pull(repository=postgis_docker_image)
        except Exception:
            test_logger.exception("Failed to pull postgres image")

        container_name = f"test_postgis_server_{uuid.uuid4()}"

        test_container = docker_client.containers.run(
            image=postgis_docker_image,
            detach=True,
            hostname="test_postgres",
            auto_remove=True,
            name=container_name,
            environment=[
                "POSTGRES_USER=postgres",
                "POSTGRES_PASSWORD=postgres",
                "POSTGRES_DB=test_volume",
            ],
            ports={"5432/tcp": 5432},
        )

        test_logger.info("STARTING IMAGE")
        try:
            time.sleep(10)
            check_database(sql_uri)
        except Exception as e:
            raise (e)
    yield
    if docker_mode:
        warnings.filterwarnings(
            action="ignore", message="unclosed", category=ResourceWarning
        )
        container = docker_client.containers.get(container_name)
        container.stop()


@pytest.fixture(scope="session")
def dynamic_annotation_interface(postgis_server, database_metadata):
    sql_uri = database_metadata["sql_uri"]
    interface = DynamicAnnotationInterface(sql_uri)
    yield interface
    interface.cached_session.close()


@pytest.fixture(scope="session")
def annotation_client(postgis_server, annotation_metadata, database_metadata):
    aligned_volume = annotation_metadata["aligned_volume"]
    sql_uri = database_metadata["sql_uri"]
    anno_client = DynamicAnnotationClient(aligned_volume, sql_uri)
    yield anno_client
    anno_client.cached_session.close()


@pytest.fixture(scope="session")
def materialization_client(postgis_server, annotation_metadata, database_metadata):
    aligned_volume = annotation_metadata["aligned_volume"]
    sql_uri = database_metadata["sql_uri"]
    mat_client = DynamicMaterializationClient(aligned_volume, sql_uri)
    yield mat_client
    mat_client.cached_session.close()


def check_database(sql_uri: str) -> None:  # pragma: no cover
    try:
        test_logger.info("ATTEMPT TO CONNECT")
        conn = psycopg2.connect(sql_uri)
        cur = conn.cursor()
        cur.execute("SELECT 1")
        test_logger.info("CONNECTED")

        cur.close()
        conn.close()
    except Exception as e:
        test_logger.info(e)



class TableNameNotFound(KeyError):
    """ Table name is not found in the Metadata table """

class TableAlreadyExists(KeyError):
    """ Table name already exists in the Metadata table """

class IdsAlreadyExists(KeyError):
    """ Annotation IDs already exists in the segmentation table """


class BadRequest(Exception):
    pass


class UpdateAnnotationError(ValueError):
    
    def __init__(self, target_id: int, superceded_id: int, message: str ="Annotation limit exceeded"):
        self.target_id = target_id
        self.message = f"""Annotation with ID {target_id} has already been superseded by annotation ID {superceded_id},
                        update annotation ID {superceded_id} instead"""
        super().__init__(self.message)

    def __str__(self):
        return f"Error update ID {self.target_id} -> {self.message}"

class AnnotationInsertLimitExceeded(ValueError):
    """Exception raised when amount of annotations exceeds defined limit.
    """

    def __init__(self, limit: int, length: int, message: str ="Annotation limit exceeded"):
        self.limit = limit
        self.message = f"The insertion limit is {limit}, {length} were attempted to be inserted"
        super().__init__(self.message)

    def __str__(self):
        return f"{self.limit} -> {self.message}"


class NoAnnotationsFoundWithID(Exception):
    """No annotation found with specified ID 
    """

    def __init__(self, anno_id: int):
        self.anno_id = anno_id
        self.message = f"No annotation with {anno_id} exists"
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}"


from sqlalchemy import Column, Boolean, String, UniqueConstraint, Integer, DateTime, Text, ForeignKey
from emannotationschemas.models import Base

class AnnoMetadata(Base):
    __tablename__ = 'annotation_table_metadata'
    id = Column(Integer, primary_key=True)
    schema_type = Column(String(100), nullable=False)
    table_name = Column(String(100), nullable=False, unique=True)
    valid = Column(Boolean)
    created = Column(DateTime, nullable=False)
    deleted = Column(DateTime, nullable=True)
    user_id = Column(String(255), nullable=False)
    description = Column(Text, nullable=False)
    reference_table = Column(String(100), nullable=True)
    flat_segmentation_source = Column(String(300), nullable=True)


class SegmentationMetadata(Base):
    __tablename__ = 'segmentation_table_metadata'
    id = Column(Integer, primary_key=True)
    schema_type = Column(String(100), nullable=False)
    table_name = Column(String(100), nullable=False, unique=True)
    valid = Column(Boolean)
    created = Column(DateTime, nullable=False)
    deleted = Column(DateTime, nullable=True)
    segmentation_source = Column(String(255), nullable=True)
    pcg_table_name = Column(String(255), nullable=False)
    last_updated = Column(DateTime, nullable=True)
    annotation_table = Column(String(100), ForeignKey('annotation_table_metadata.table_name'))



import logging

from sqlalchemy import Table


def test_create_or_select_database(dynamic_annotation_interface,
                                   database_metadata,
                                   annotation_metadata):
    aligned_volume = annotation_metadata["aligned_volume"]
    sql_uri = database_metadata["sql_uri"]
    new_sql_uri = dynamic_annotation_interface.create_or_select_database(
        aligned_volume, sql_uri
    )
    logging.info(new_sql_uri)
    assert str(new_sql_uri) == database_metadata["sql_uri"]


def test_get_table_metadata(dynamic_annotation_interface, annotation_metadata):
    table_name = annotation_metadata["table_name"]
    schema_type = annotation_metadata["schema_type"]
    metadata = dynamic_annotation_interface.get_table_metadata(table_name)
    logging.info(metadata)
    assert metadata["schema_type"] == schema_type
    assert metadata["table_name"] == "anno_test"
    assert metadata["user_id"] == "foo@bar.com"
    assert metadata["description"] == "some description"


def test_get_table_schema(dynamic_annotation_interface, annotation_metadata):
    table_name = annotation_metadata["table_name"]
    schema_type = annotation_metadata["schema_type"]
    schema_info = dynamic_annotation_interface.get_table_schema(table_name)
    assert schema_info == schema_type


def test_get_table_sql_metadata(dynamic_annotation_interface, annotation_metadata):
    table_name = annotation_metadata["table_name"]

    sql_metadata = dynamic_annotation_interface.get_table_sql_metadata(table_name)
    logging.info(sql_metadata)
    assert isinstance(sql_metadata, Table)


def test_get_model_columns(dynamic_annotation_interface, annotation_metadata):
    table_name = annotation_metadata["table_name"]

    model_columns = dynamic_annotation_interface._get_model_columns(table_name)
    logging.info(model_columns)
    assert isinstance(model_columns, list)


def test_get_existing_table_ids(dynamic_annotation_interface):
    table_names = dynamic_annotation_interface._get_existing_table_names()
    assert isinstance(table_names, list)


def test_has_table(dynamic_annotation_interface, annotation_metadata):
    table_name = annotation_metadata["table_name"]

    has_table = dynamic_annotation_interface.has_table(table_name)
    assert has_table == True


def test_get_annotation_table_size(dynamic_annotation_interface, annotation_metadata):
    table_name = annotation_metadata["table_name"]

    table_size = dynamic_annotation_interface.get_annotation_table_size(table_name)
    assert table_size == 3



import logging


def test_create_table(annotation_client, annotation_metadata):
    table_name = annotation_metadata["table_name"]
    schema_type = annotation_metadata["schema_type"]

    table = annotation_client.create_table(
        table_name,
        schema_type,
        description="some description",
        user_id="foo@bar.com",
        reference_table=None,
        flat_segmentation_source=None,
    )
    assert table_name == table


def test_insert_annotation(annotation_client, annotation_metadata):
    table_name = annotation_metadata["table_name"]

    test_data = [
        {
            "pre_pt": {"position": [121, 123, 1232]},
            "ctr_pt": {"position": [121, 123, 1232]},
            "post_pt": {"position": [333, 555, 5555]},
            "size": 1,
        }
    ]
    is_commmited = annotation_client.insert_annotations(table_name, test_data)
    annotation_client.cached_session.close()

    assert is_commmited == True


def test_insert_another_annotation(annotation_client, annotation_metadata):
    table_name = annotation_metadata["table_name"]

    test_data = [
        {
            "pre_pt": {"position": [111, 222, 333]},
            "ctr_pt": {"position": [444, 555, 666]},
            "post_pt": {"position": [777, 888, 999]},
            "size": 1,
        }
    ]
    is_commmited = annotation_client.insert_annotations(table_name, test_data)
    annotation_client.cached_session.close()

    assert is_commmited == True


def test_get_annotation(annotation_client, annotation_metadata):
    table_name = annotation_metadata["table_name"]
    test_data = annotation_client.get_annotations(table_name, [1])
    logging.info(test_data)
    annotation_client.cached_session.close()

    assert test_data[0]["id"] == 1


def test_update_annotation(annotation_client, annotation_metadata):
    table_name = annotation_metadata["table_name"]

    updated_test_data = {
        "id": 1,
        "pre_pt": {"position": [222, 123, 1232]},
        "ctr_pt": {"position": [121, 123, 1232]},
        "post_pt": {"position": [555, 555, 5555]},
    }
    is_updated = annotation_client.update_annotation(table_name, updated_test_data)
    annotation_client.cached_session.close()

    assert is_updated == "id 1 updated"


def test_delete_annotation(annotation_client, annotation_metadata):
    table_name = annotation_metadata["table_name"]

    ids_to_delete = [1]
    is_deleted = annotation_client.delete_annotation(table_name, ids_to_delete)
    annotation_client.cached_session.close()

    assert is_deleted == True



from dynamicannotationdb.models import SegmentationMetadata
from dynamicannotationdb.interface import DynamicAnnotationInterface
from dynamicannotationdb.errors import AnnotationInsertLimitExceeded, UpdateAnnotationError, IdsAlreadyExists
from emannotationschemas import get_schema, get_flat_schema
from emannotationschemas.flatten import flatten_dict
from emannotationschemas import models as em_models
from dynamicannotationdb.key_utils import build_segmentation_table_name
from marshmallow import INCLUDE, EXCLUDE
from sqlalchemy.exc import ArgumentError, InvalidRequestError, OperationalError, IntegrityError
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm.exc import NoResultFound
from typing import List
import logging
import datetime
import json


class DynamicMaterializationClient(DynamicAnnotationInterface):
    def __init__(self, aligned_volume, sql_base_uri):
        sql_uri = self.create_or_select_database(aligned_volume, sql_base_uri)
        super().__init__(sql_uri)
    
        self.aligned_volume = aligned_volume
        self._table = None
        self._cached_schemas = {}

    @property
    def table(self):
        return self._table

    def load_table(self, table_name: str):
        self._table = self._cached_table(table_name)
        return self._table


    def create_and_attach_seg_table(self, table_name: str,
                                          pcg_table_name: str):
                
        schema_type = self.get_table_schema(table_name)
        return self.create_segmentation_table(table_name,
                                              schema_type,
                                              pcg_table_name)

    def drop_table(self, table_name: str) -> bool:
        return self._drop_table(table_name)

    def get_linked_tables(self, table_name: str, pcg_table_name: str) -> List:
        try:
            linked_tables = self.cached_session.query(SegmentationMetadata).\
                            filter(SegmentationMetadata.annotation_table==table_name).\
                            filter(SegmentationMetadata.pcg_table_name==pcg_table_name).all()    
            return linked_tables
        except Exception as e:
            raise AttributeError(f"No table found with name '{table_name}'. Error: {e}")        

    def get_linked_annotations(self, table_name: str,
                                     pcg_table_name: str,
                                     annotation_ids: List[int]) -> dict:
        """ Get list of annotations from database by id.

        Parameters
        ----------
        table_name : str
            name of annotation table
        pcg_table_name: str
            name of chunked graph reference table
        annotation_ids : int
            annotation id 

        Returns
        -------
        list
            list of annotation data dicts
        """
        
        schema_type = self.get_table_schema(table_name)
        
        seg_table_name = build_segmentation_table_name(table_name, pcg_table_name)
        AnnotationModel = self._cached_table(table_name)
        SegmentationModel = self._cached_table(seg_table_name)
        
        annotations = self.cached_session.query(AnnotationModel, SegmentationModel).\
                                          join(SegmentationModel, SegmentationModel.id==AnnotationModel.id).\
                                          filter(AnnotationModel.id.in_([x for x in annotation_ids])).all()

        
        FlatSchema = get_flat_schema(schema_type)
        schema = FlatSchema(unknown=INCLUDE)
        
        data = []
        for anno, seg in annotations: 
            anno_data = anno.__dict__
            seg_data = seg.__dict__
            anno_data['created'] = str(anno_data.get('created'))
            anno_data['deleted'] = str(anno_data.get('deleted'))
            anno_data.pop('_sa_instance_state', None)
            seg_data.pop('_sa_instance_state', None)
            merged_data = {**anno_data, **seg_data}
            data.append(merged_data)

        return schema.load(data, many=True)

    def insert_linked_segmentation(self, table_name:str,
                                         pcg_table_name: str,
                                         segmentations: List[dict]):
        """Insert segmentations by linking to annotation ids. Limited to 10,000 segmentations. 
        If more consider using a bulk insert script.

        Parameters
        ----------
        table_name : str
            name of annotation table
        pcg_table_name: str
            name of chunked graph reference table
        segmentations : List[dict]
            List of dictionaries of single segmentation data. 
        """
        insertion_limit = 10_000

        if len(segmentations) > insertion_limit:
            raise AnnotationInsertLimitExceeded(len(segmentations), insertion_limit)
                
        schema_type = self.get_table_schema(table_name)
        seg_table_name = build_segmentation_table_name(table_name, pcg_table_name)

        SegmentationModel = self._cached_table(seg_table_name)
        formatted_seg_data = []

        _, segmentation_schema = self._get_flattened_schema(schema_type)

        for segmentation in segmentations:
            segmentation_data = flatten_dict(segmentation)
            flat_data = self._map_values_to_schema(segmentation_data, segmentation_schema)
            flat_data['id'] = segmentation['id']

            formatted_seg_data.append(flat_data)

        segs = [SegmentationModel(**segmentation_data)
                        for segmentation_data in formatted_seg_data]
        
        ids = [data['id'] for data in formatted_seg_data]
        q = self.cached_session.query(SegmentationModel).filter(SegmentationModel.id.in_([id for id in ids]))
        ids_exist = self.cached_session.query(q.exists()).scalar() 
        
        if not ids_exist: # TODO replace this with a filter for ids that are missing from this list
            self.cached_session.add_all(segs)
            self.commit_session()
            return True
        else:
            raise IdsAlreadyExists(f"Annotation IDs {ids} already linked in database ")

    def insert_linked_annotations(self, table_name:str,
                                        pcg_table_name: str,
                                        annotations: List[dict]):
        """Insert annotations by type and schema. Limited to 10,000 annotations. If more consider
        using a bulk insert script.

        Parameters
        ----------
        table_name : str
            name of annotation table
        pcg_table_name: str
            name of chunked graph reference table
        annotations : dict
            Dictionary of single annotation data. 
        """
        insertion_limit = 10_000

        if len(annotations) > insertion_limit:
            raise AnnotationInsertLimitExceeded(len(annotations), insertion_limit)
        
        schema_type = self.get_table_schema(table_name)

        seg_table_name = build_segmentation_table_name(table_name, pcg_table_name)

        formatted_anno_data = []
        formatted_seg_data = []
        
        AnnotationModel = self._cached_table(table_name)
        SegmentationModel = self._cached_table(seg_table_name)
        
        for annotation in annotations:
            
            annotation_data, segmentation_data = self._get_flattened_schema_data(
                schema_type, annotation)
            if annotation.get('id'):
                annotation_data['id'] = annotation['id']
                
            annotation_data['created'] = datetime.datetime.now()
            
            formatted_anno_data.append(annotation_data)
            formatted_seg_data.append(segmentation_data)
            
        annos = [AnnotationModel(**annotation_data)
                 for annotation_data in formatted_anno_data]

        self.cached_session.add_all(annos)
        self.cached_session.flush()
        segs = [SegmentationModel(**segmentation_data, id=anno.id)
                for segmentation_data, anno in zip(formatted_seg_data, annos)]
        self.cached_session.add_all(segs)
        self.commit_session()
        return True

    def update_linked_annotations(self, table_name: str,
                                        pcg_table_name: str,
                                        annotation: dict):
        """Updates an annotation by inserting a new row. The original annotation
        will refer to the new row with a superceded_id. Does not update inplace.

        Parameters
        ----------
        table_name : str
            name of annotation table
        pcg_table_name: str
            name of chunked graph reference table
        annotation : dict, annotation to update by ID
        """
        anno_id = annotation.get('id')
        if not anno_id:
            return "Annotation requires an 'id' to update targeted row"

        seg_table_name = build_segmentation_table_name(table_name, pcg_table_name)

        schema_type = self.get_table_schema(table_name)

        AnnotationModel = self._cached_table(table_name)
        SegmentationModel = self._cached_table(seg_table_name)
        
        new_annotation, __ = self._get_flattened_schema_data(schema_type, annotation)
        
        new_annotation['created'] = datetime.datetime.now()
        new_annotation['valid'] = True

        new_data = AnnotationModel(**new_annotation)
        try:
            data = self.cached_session.query(AnnotationModel, SegmentationModel).filter(AnnotationModel.id==anno_id).filter(SegmentationModel.id==anno_id).all()
            for old_anno, old_seg in data:
                if old_anno.superceded_id:
                    raise UpdateAnnotationError(anno_id, old_anno.superceded_id)
                
                self.cached_session.add(new_data)
                self.cached_session.flush()
                
                deleted_time = datetime.datetime.now()
                old_anno.deleted = deleted_time
                old_anno.superceded_id = new_data.id
                old_anno.valid = False
                                    
                self.commit_session()
                return f"id {anno_id} updated"
        except NoResultFound as e:
            return f"No result found for {anno_id}. Error: {e}"

    def delete_linked_annotation(self, table_name: str,
                                       pcg_table_name: str,
                                       annotation_ids: List[int]):
        """Mark annotations by for deletion by list of ids. 

        Parameters
        ----------
        table_name : str
            name of annotation table
        pcg_table_name: str
            name of chunked graph reference table
        annotation_ids : List[int]
            list of ids to delete

        Returns
        -------

        Raises
        ------
        """
        seg_table_name = build_segmentation_table_name(table_name, pcg_table_name)
        AnnotationModel = self._cached_table(table_name)
        SegmentationModel = self._cached_table(seg_table_name)
        
        annotations = self.cached_session.query(AnnotationModel).\
                                          join(SegmentationModel, SegmentationModel.id==AnnotationModel.id).\
                                          filter(AnnotationModel.id.in_([x for x in annotation_ids])).all()

        if annotations:
            deleted_time = datetime.datetime.now()
            for annotation in annotations:
                annotation.deleted = deleted_time
                annotation.valid = False
            self.commit_session()
        else:
            return None
        return True



import logging

from sqlalchemy.ext.declarative import api


def test_load_table(materialization_client, annotation_metadata):
    table_name = annotation_metadata["table_name"]

    loaded_table = materialization_client.load_table(table_name)
    if isinstance(loaded_table, api.DeclarativeMeta):
        assert loaded_table.__name__ == table_name


def test_create_and_attach_seg_table(materialization_client, annotation_metadata):
    table_name = annotation_metadata["table_name"]
    pcg_table_name = annotation_metadata["pcg_table_name"]

    table_added_status = materialization_client.create_and_attach_seg_table(
        table_name, pcg_table_name
    )
    assert table_added_status == {
        "Created Succesfully": True,
        "Table Name": f"{table_name}__{pcg_table_name}",
    }


def test_insert_linked_annotations(materialization_client, annotation_metadata):
    table_name = annotation_metadata["table_name"]
    pcg_table_name = annotation_metadata["pcg_table_name"]

    segmentation_data = [
        {
            "id": 8,
            "pre_pt": {
                "position": [121, 123, 1232],
                "supervoxel_id": 2344444,
                "root_id": 4,
            },
            "ctr_pt": {"position": [121, 123, 1232]},
            "post_pt": {
                "position": [121, 123, 1232],
                "supervoxel_id": 3242424,
                "root_id": 5,
            },
            "size": 2,
        }
    ]

    is_inserted = materialization_client.insert_linked_annotations(
        table_name, pcg_table_name, segmentation_data
    )
    materialization_client.cached_session.close()
    assert is_inserted == True


def test_get_linked_annotations(materialization_client, annotation_metadata):
    table_name = annotation_metadata["table_name"]
    pcg_table_name = annotation_metadata["pcg_table_name"]

    annotations = materialization_client.get_linked_annotations(
        table_name, pcg_table_name, [8]
    )
    materialization_client.cached_session.close()

    logging.info(annotations)
    assert annotations[0]["pre_pt_supervoxel_id"] == 2344444
    assert annotations[0]["pre_pt_root_id"] == 4
    assert annotations[0]["post_pt_supervoxel_id"] == 3242424
    assert annotations[0]["post_pt_root_id"] == 5



def build_table_id(aligned_volume: str, table_name: str) -> str:
    """Combines aligned_volume name and specified table name to 
    create specific table id


    Parameters
    ----------
    aligned_volume : str
        name of aligned volume
    table_name : str
        name of table assigned to an aligned volume

    Returns
    -------
    str
        formatted table_id 
    """
    return f"{aligned_volume}__{table_name}"

def build_segmentation_table_name(annotation_table_name: str, 
                                segmentation_source: str) -> str:
    """Creates a table name that combines annotation table and appends 
    segmentation table name

    Parameters
    ----------
    annotation_table_name : str
        exiting annotation table name
    pcg_table_name : str
        name of pychunkedgraph table

    Returns
    -------
    str
        formatted name of table combining the annotation table id with 
        pychunkedgraph segmentation source name
    """
    return f"{annotation_table_name}__{segmentation_source}"

def get_table_name_from_table_id(table_id: str) -> str:
    """Extracts table name from table_id string
    Parameters
    ----------
    table_id : str

    Returns
    -------
    str
        table name in table id
    """
    return table_id.split("__")[-1]


def get_dataset_name_from_table_id(table_id: str) -> str:
    """Extracts the aligned volume name from table id string

    Parameters
    ----------
    table_id : str

    Returns
    -------
    str
        name of aligned volume in table id
    """
    return table_id.split("__")[1]


from sqlalchemy import create_engine, inspect, func, MetaData
from sqlalchemy.orm import sessionmaker, Session, scoped_session
from sqlalchemy.exc import ArgumentError, InvalidRequestError, OperationalError, IntegrityError
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.engine.url import make_url
from sqlalchemy.ext.declarative.api import DeclarativeMeta
from sqlalchemy.pool import NullPool
from marshmallow import EXCLUDE
from emannotationschemas import get_schema, get_flat_schema
from emannotationschemas.flatten import flatten_dict
from emannotationschemas import models as em_models
from dynamicannotationdb.key_utils import build_segmentation_table_name
from dynamicannotationdb.models import AnnoMetadata, SegmentationMetadata
from dynamicannotationdb.errors import TableNameNotFound, TableAlreadyExists
from typing import List
import logging
import datetime
import time


class DynamicAnnotationInterface:

    def __init__(self, sql_uri: str):
        """ Annotation DB interface layer for creating and querying tables in SQL

        Parameters
        ----------
        sql_uri : str
            SQL URI to use to connect to a Database.
        create_metadata : bool, optional
            Creates additional columns on new tables for CRUD operations, by default True
        """
        self.sql_uri = sql_uri
        self.engine = create_engine(sql_uri,
                               pool_recycle=3600,
                               pool_size=20,
                               max_overflow=50)
        self.base = em_models.Base
        self.base.metadata.bind = self.engine

        table_objects = [
            AnnoMetadata.__tablename__,
            SegmentationMetadata.__tablename__,
        ]
        for table in table_objects:
            if not self.engine.dialect.has_table(self.engine, table): 
                self.base.metadata.tables[table].create(bind=self.engine)

        self.mapped_base = automap_base()
        self.mapped_base.prepare(self.engine, reflect=True)
        
        self.session = scoped_session(sessionmaker(bind=self.engine, autocommit=False, autoflush=False))

        self.insp = inspect(self.engine)

        self._cached_session = None
        self._cached_tables = {}

    def create_or_select_database(self, aligned_volume: str, sql_uri: str):
        """Create a new database with the name of the aligned volume. Checks if 
        database exists before creating.

        Parameters
        ----------
        aligned_volume : str
            name of aligned volume which the database name will inherent 
        sql_uri : str
            base path to the sql server

        Returns
        -------
        sql_url instance
        """
        sql_base_uri = sql_uri.rpartition("/")[0]

        sql_uri = make_url(f"{sql_base_uri}/{aligned_volume}")
        
        temp_engine = create_engine(
            sql_base_uri, poolclass=NullPool, isolation_level='AUTOCOMMIT', pool_pre_ping=True)

        with temp_engine.connect() as connection:
            connection.execute("commit")
            database_exists = connection.execute(
                f"SELECT 1 FROM pg_catalog.pg_database WHERE datname = '{sql_uri.database}'")
            if not database_exists.fetchone():
                logging.info(f"Creating new database: {sql_uri.database}")

                connection.execute(f"SELECT pg_terminate_backend(pid) FROM pg_stat_activity \
                           WHERE pid <> pg_backend_pid() AND datname = '{sql_uri.database}';")

                # check if template exists, create if missing
                template_exist = connection.execute(
                    f"SELECT 1 FROM pg_catalog.pg_database WHERE datname = 'template_postgis'")

                if not template_exist.fetchone():

                    # create postgis template db
                    connection.execute(f"CREATE DATABASE template_postgis")

                    # create postgis extension
                    template_uri = make_url(f"{sql_base_uri}/template_postgis")
                    tempate_engine = create_engine(
                        template_uri, poolclass=NullPool, isolation_level='AUTOCOMMIT', pool_pre_ping=True)
                    with tempate_engine.connect() as template_connection:
                        template_connection.execute(
                            'CREATE EXTENSION IF NOT EXISTS postgis')
                    tempate_engine.dispose()

                # finally create new annotation database
                connection.execute(
                    f"CREATE DATABASE {sql_uri.database} TEMPLATE template_postgis")

        temp_engine.dispose()
        return sql_uri


    @property
    def cached_session(self) -> Session:
        if self._cached_session is None:
            self._cached_session = self.session()
        return self._cached_session

    def commit_session(self):
        try:
            self.cached_session.commit()
        except Exception:
            self.cached_session.rollback()
            logging.exception(f"SQL Error")
        finally:
            self.cached_session.close()
        self._cached_session = None

    def create_annotation_table(self,
                                table_name: str,
                                schema_type:str,
                                description: str,
                                user_id: str,
                                reference_table: str = None,
                                flat_segmentation_source: str=None):
        r"""Create new annotation table unless already exists

        Parameters
        ----------
        aligned_volume : str
            name of aligned_volume to attach a new annotation table
        table_name : str
            name of table
        schema_type : str
            Type of schema to use, must be a valid type from EMAnnotationSchemas

         description: str
            a string with a human readable explanation of 
            what is in the table. Including who made it
            and any information that helps interpret the fields
            of the annotations.

        user_id: str
            user id for this table

        reference_table: str
            reference table name, if required by this schema

        flat_segmentation_source: str
            a path to a segmentation source associated with this table
             i.e. 'precomputed:\\gs:\\my_synapse_seg\example1'
        """

        if table_name in self._get_existing_table_names():
            logging.warning(f"Table creation failed: {table_name} already exists")
            raise TableAlreadyExists
        
        model = em_models.make_annotation_model(table_name,
                                                schema_type)

        self.base.metadata.tables[model.__name__].create(bind=self.engine)
        creation_time = datetime.datetime.now()
        
        metadata_dict = {
            'description': description,
            'user_id': user_id,
            'reference_table': reference_table,
            'schema_type': schema_type,
            'table_name': table_name,
            'valid': True,
            'created': creation_time,
            'flat_segmentation_source': flat_segmentation_source
        }
        logging.info(f"Metadata for table: {table_name} is {metadata_dict}")
        anno_metadata = AnnoMetadata(**metadata_dict)
        self.cached_session.add(anno_metadata)
        self.commit_session()
        logging.info(f"Table: {table_name} created using {model} model at {creation_time}")
        return table_name

    def create_segmentation_table(self, annotation_table_name: str,
                                        schema_type:str,
                                        pcg_table_name: str):
        """Create new segmentation table linked to an annotation table
         unless it already exists. 

        Parameters
        ----------
        annotation_table_name : str
            name of table annotation to link the segmentation table 
        schema_type : str
            Type of schema to use, must be a valid type from EMAnnotationSchemas
        pcg_table_name : str
            name of pychunked graph segmentation to use

        Returns
        -------
        dict
            description of table that is created
        """

        segmentation_table_name = build_segmentation_table_name(annotation_table_name,
                                                            pcg_table_name)
        
        if annotation_table_name not in self._get_existing_table_names():
            raise TableNameNotFound

        model = em_models.make_segmentation_model(annotation_table_name,
                                                    schema_type,
                                                    pcg_table_name)

        self.base.metadata.tables[model.__name__].create(bind=self.engine)
        creation_time = datetime.datetime.now()

        metadata_dict = {
            'annotation_table': annotation_table_name,
            'schema_type': schema_type,
            'table_name': segmentation_table_name,
            'valid': True,
            'created': creation_time,
            'pcg_table_name': pcg_table_name
        }
        seg_metadata = SegmentationMetadata(**metadata_dict)
        self.cached_session.add(seg_metadata)           
        self.commit_session()

        logging.info(f"Table: {segmentation_table_name} created using {model} model at {creation_time}")
        return {"Created Succesfully": True, "Table Name": model.__name__}

    def get_table_metadata(self, table_name: str):
        metadata = self.cached_session.query(AnnoMetadata).\
                        filter(AnnoMetadata.table_name==table_name).first()
        try:
            metadata.__dict__.pop('_sa_instance_state')
            return metadata.__dict__
        except Exception as e:
            raise AttributeError(f"No table found with name '{table_name}'. Error: {e}")

    def get_segmentation_table_metadata(self, table_name: str, pcg_table_name: str):
        seg_table_name = build_segmentation_table_name(table_name, pcg_table_name)
        metadata = self.cached_session.query(SegmentationMetadata).\
                        filter(SegmentationMetadata.table_name==seg_table_name).first()
        try:
            metadata.__dict__.pop('_sa_instance_state')
            return metadata.__dict__
        except Exception as e:
            raise AttributeError(f"No table found with name '{table_name}'. Error: {e}")

    def _get_existing_table_by_name(self) -> List[str]:
        """Get the table names of table that exist

        Returns
        -------
        List[str]
            list of table names that exist
        """
        table_names = self._get_existing_table_names()
        return table_names

    def _get_existing_table_names_metadata(self) -> List[dict]:
        """Get all the metadata for all tables

        Returns
        -------
        List[dict]
            all table metadata that exist
        """
        return [
            self.get_table_metadata(table_name)
            for table_name in self._get_existing_table_names()
        ]

    def get_table_schema(self, table_name: str):
        table_metadata = self.get_table_metadata(table_name)
        return table_metadata['schema_type']

    def get_table_sql_metadata(self, table_name):
        self.base.metadata.reflect(bind=self.engine)
        return self.base.metadata.tables[table_name]

    def _get_model_columns(self, table_name: str) -> list:
        """ Return list of column names and types of a given table

        Parameters
        ----------
        table_name : str
            Table name in database

        Returns
        -------
        list
            column names and types
        """
        try:
            model_columns = []
            db_columns = self.insp.get_columns(table_name)
            for column in db_columns:
                model_columns.append(tuple([column['name'],
                                            column['type']]))
            return model_columns
        except Exception as e:
            raise TableNameNotFound(f"Error: {e}. No table name exists with name {table_name}.")

    def _get_all_tables(self):
        return self.cached_session.query(AnnoMetadata).all()

    def _get_existing_table_names(self):
        """ Collects table_names keys of existing tables

        Returns
        -------
        list
            List of table_names
        """
        metadata = self.cached_session.query(AnnoMetadata).all()
        return [m.table_name for m in metadata]

    def get_valid_table_names(self):
        metadata = self.cached_session.query(AnnoMetadata).all()
        return [m.table_name for m in metadata if m.valid==True]

    def get_existing_segmentation_table_names(self):
        """ Collects table_names keys of existing segmentation tables
        contained in an aligned volume database. Used for materialization.

        Returns
        -------
        list
            List of segmentation table_names
        """
        metadata = self.cached_session.query(SegmentationMetadata).all()
        return [m.table_name for m in metadata]

    def has_table(self, table_name: str) -> bool:
        """Check if a table exists

        Parameters
        ----------
        table_name : str
            Table name in database

        Returns
        -------
        bool
            whether the table exists
        """
        return table_name in self._get_existing_table_names()

    def _cached_table(self, table_name: str) -> DeclarativeMeta:
        """ Returns cached table 'DeclarativeMeta' callable for querying.

        Parameters
        ----------
        table_name : str
            Table name in database
        Returns
        -------
        DeclarativeMeta
            SQLAlchemy callable. See 'https://docs.sqlalchemy.org/en/13/orm/extensions/declarative/api.html#sqlalchemy.ext.declarative.declarative_base.params.metaclass'
        """
        try:
            self._load_table(table_name)
            return self._cached_tables[table_name]
        except TableNameNotFound as error:
            logging.error(f"Cannot load table {error}")

    def _get_table_row_count(self, table_name: str, filter_valid: bool=False) -> int:
        model = self._cached_table(table_name)
        if filter_valid:
            row_count = self.cached_session.query(func.count(model.id)).filter(model.valid==True).scalar()
        else:
            row_count = self.cached_session.query(func.count(model.id)).scalar()
        return row_count

    def get_max_id_value(self, table_name: str) -> int:
        model = self._cached_table(table_name)
        return self.cached_session.query(func.max(model.id)).scalar()

    def get_min_id_value(self, table_name: str) -> int:
        model = self._cached_table(table_name)
        return self.cached_session.query(func.min(model.id)).scalar()

    def get_annotation_table_size(self, table_name: str) -> int:
        """Get the number of annotations in a table

        Parameters
        ----------
        table_name : str
            name of table contained within the aligned_volume database

        Returns
        -------
        int
            number of annotations
        """
        Model = self._cached_table(table_name)
        return self.cached_session.query(Model).count()

    def _drop_table(self, table_name: str):
        table = self.base.metadata.tables.get(table_name)
        if table:
            logging.info(f'Deleting {table_name} table')
            self.base.metadata.drop_all(self.engine, [table], checkfirst=True)
            if self._is_cached(table):
                del self._cached_tables[table]
            return True
        return False

    def get_annotation_model(self, table_name: str):
        return self._get_model_from_table_name(table_name)

    def get_segmentation_model(self, table_name: str, pcg_table_name: str):
        table_name = build_segmentation_table_name(table_name, pcg_table_name)
        return self._get_model_from_table_name(table_name)

    def _get_model_from_table_name(self, table_name: str) -> DeclarativeMeta:
        self.mapped_base = automap_base()
        self.mapped_base.prepare(self.engine, reflect=True)
        return self.mapped_base.classes[table_name]

    def _get_flattened_schema_data(self, schema_type: str, data: dict) -> dict:
        schema_type = get_schema(schema_type)
        schema = schema_type(context={'postgis': True})
        data = schema.load(data, unknown=EXCLUDE)

        check_is_nested = (any(isinstance(i, dict) for i in data.values()))
        if check_is_nested:
            data = flatten_dict(data)

        flat_annotation_schema, flat_segmentation_schema = em_models.split_annotation_schema(schema_type)

        return self._map_values_to_schema(data, flat_annotation_schema), self._map_values_to_schema(data, flat_segmentation_schema)

    def _get_flattened_schema(self, schema_type: str):
        schema_type = get_schema(schema_type)

        flat_annotation_schema, flat_segmentation_schema = em_models.split_annotation_schema(schema_type)

        return flat_annotation_schema, flat_segmentation_schema

    def _map_values_to_schema(self, data, schema):
        return {key: data[key] for key, value in schema._declared_fields.items() if key in data}

    def _is_cached(self, table_name: str) -> bool:
        """ Check if table is loaded into cached instance dict of tables

        Parameters
        ----------
        table_name : str
            Name of table to check if loaded

        Returns
        -------
        bool
            True if table is loaded else False.
        """

        return table_name in self._cached_tables

    def _load_table(self, table_name: str):
        """ Load existing table into cached lookup dict instance

        Parameters
        ----------
        table_name : str
            Table name to be loaded from existing database tables

        Returns
        -------
        bool
            Returns True if table exists and is loaded into cached table dict.
        """
        if self._is_cached(table_name):
            return True

        try:
            self._cached_tables[table_name] = self._get_model_from_table_name(table_name)
            return True
        except KeyError as key_error:
            if table_name in self._get_existing_table_names():
                logging.error(f"Could not load table: {key_error}")
            return False





from setuptools import setup, find_packages
import os
import codecs
import re

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as f:
    required = f.read().splitlines()

setup(
    name="DynamicAnnotationDB",
    version=find_version("dynamicannotationdb", "__init__.py"),
    author="Sven Dorkenwald, Derrick Brittain",
    author_email="sdorkenw@princeton.edu",
    description="Annotation Database pendant to the chunkedgraph",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/seung-lab/DynamicAnnotationDB",
    packages=find_packages(),
    install_requires=required,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ),
)



from dynamicannotationdb.interface import DynamicAnnotationInterface
from dynamicannotationdb.errors import AnnotationInsertLimitExceeded, \
                                      NoAnnotationsFoundWithID, \
                                      UpdateAnnotationError
from dynamicannotationdb.models import AnnoMetadata
from emannotationschemas import get_flat_schema
from marshmallow import INCLUDE
from sqlalchemy.exc import ArgumentError, InvalidRequestError, OperationalError, IntegrityError
from sqlalchemy.orm.exc import NoResultFound

from typing import List
import datetime
import logging
import json


class DynamicAnnotationClient(DynamicAnnotationInterface):
    
    def __init__(self, aligned_volume, sql_base_uri):
        sql_uri = self.create_or_select_database(aligned_volume, sql_base_uri)
        super().__init__(sql_uri)

        self.aligned_volume = aligned_volume

        self._table = None
        self._cached_schemas = {}

    @property
    def table(self):
        return self._table

    def load_table(self, table_name: str):
        """Load a table

        Parameters
        ----------
        table_name : str
            name of table

        Returns
        -------
        DeclarativeMeta
            the sqlalchemy table of that name
        """
        self._table = self._cached_table(table_name)
        return self._table

    def create_table(self, table_name: str,
                     schema_type: str,
                     description: str,
                     user_id: str,
                     reference_table: str = None,
                     flat_segmentation_source: str = None):
        r"""Create a new annotation table

        Parameters
        ----------
        table_name : str
            name of new table
            
        schema_type : str
            type of schema for that table

        description: str
            a string with a human readable explanation of 
            what is in the table. Including who made it
            and any information that helps interpret the fields
            of the annotations.

        user_id: str
            user id for this table

        reference_table: str
            reference table name, if required by this schema

        flat_segmentation_source: str
            a path to a segmentation source associated with this table
             i.e. 'precomputed:\\gs:\\my_synapse_seg\example1'

        Returns
        -------
        str
           desciption of table at creation time
        """
        # TODO: check that schemas that are reference schemas
        # have a reference_table in their metadata
        return self.create_annotation_table(table_name,
                                            schema_type,
                                            description,
                                            user_id,
                                            reference_table=reference_table,
                                            flat_segmentation_source=flat_segmentation_source)

    def delete_table(self, table_name: str) -> bool:
        """Marks a table for deletion, which will
        remove it from user visible calls
        and stop materialization from happening on this table
        only updates metadata to reflect deleted timestamp.

        Parameters
        ----------
        table_name : str
             name of table to mark for deletion

        Returns
        -------
        bool
            whether table was successfully deleted
        """
        metadata = self.cached_session.query(AnnoMetadata). \
            filter(AnnoMetadata.table_name == table_name).first()
        metadata.deleted = datetime.datetime.now()
        self.commit_session()
        return True

    def drop_table(self, table_name: str) -> bool:
        """Drop a table, actually removes it from the database
        along with segmentation tables associated with it

        Parameters
        ----------
        table_name : str
            name of table to drop

        Returns
        -------
        bool
            whether drop was successful
        """
        return self._drop_table(table_name)

    def insert_annotations(self, table_name: str,
                          annotations: List[dict]):
        """Insert some annotations.

        Parameters
        ----------
        table_name : str
            name of target table to insert annotations
        annotations : list of dict
            a list of dicts with the annotations
                                that meet the schema

        Returns
        -------
        bool
            True is succesfully inserted annotations

        Raises
        ------
        AnnotationInsertLimitExceeded
            Exception raised when amount of annotations exceeds defined limit.
        """
        insertion_limit = 10_000

        if len(annotations) > insertion_limit:
            raise AnnotationInsertLimitExceeded(len(annotations), insertion_limit)

        schema_type = self.get_table_schema(table_name)

        AnnotationModel = self._cached_table(table_name)

        formatted_anno_data = []
        for annotation in annotations:

            annotation_data, __ = self._get_flattened_schema_data(schema_type, annotation)
            if annotation.get('id'):
                annotation_data['id'] = annotation['id']

            annotation_data['created'] = datetime.datetime.now()
            annotation_data['valid'] = True
            formatted_anno_data.append(annotation_data)

        annos = [AnnotationModel(**annotation_data) for annotation_data in formatted_anno_data]

        self.cached_session.add_all(annos)
        self.commit_session()
        return True

    def get_annotations(self, table_name: str,
                       annotation_ids: List[int]) -> List[dict]:
        """Get a set of annotations by ID

        Parameters
        ----------
        table_name : str
            name of table
        annotation_ids : List[int]
            list of annotation ids to get

        Returns
        -------
        List[dict]
            list of returned annotations
        """
        AnnotationModel = self._cached_table(table_name)

        annotations = self.cached_session.query(AnnotationModel). \
            filter(AnnotationModel.id.in_([x for x in annotation_ids])).all()
        
        schema_type = self.get_table_schema(table_name)
        anno_schema, __ = self._get_flattened_schema(schema_type)
        schema = anno_schema(unknown=INCLUDE)
        try:
            data = []

            for anno in annotations:
                anno_data = anno.__dict__
                anno_data['created'] = str(anno_data.get('created'))
                anno_data['deleted'] = str(anno_data.get('deleted'))
                anno_data.pop('_sa_instance_state', None)
                data.append(anno_data)

            return schema.load(data, many=True)

        except Exception as e:
            logging.exception(e)
            raise NoAnnotationsFoundWithID(annotation_ids)

    def update_annotation(self, table_name: str,
                          annotation: dict):
        """Update an annotation

        Parameters
        ----------
        table_name : str
            name of targeted table to update annotations
        annotation : dict
            new data for that annotation

        Returns
        -------
        [type]
            [description]

        Raises
        ------
        TODO: make this raise an exception rather than return strings
        """
        anno_id = annotation.get('id')
        if not anno_id:
            return "Annotation requires an 'id' to update targeted row"
        schema_type = self.get_table_schema(table_name)

        AnnotationModel = self._cached_table(table_name)

        new_annotation, __ = self._get_flattened_schema_data(schema_type, annotation)

        new_annotation['created'] = datetime.datetime.now()
        new_annotation['valid'] = True

        new_data = AnnotationModel(**new_annotation)
        try:
            old_anno = self.cached_session.query(AnnotationModel).filter(AnnotationModel.id == anno_id).one()

            if old_anno.superceded_id:
                raise UpdateAnnotationError(anno_id, old_anno.superceded_id)

            self.cached_session.add(new_data)
            self.cached_session.flush()
            
            deleted_time = datetime.datetime.now()
            old_anno.deleted = deleted_time
            old_anno.superceded_id = new_data.id
            old_anno.valid = False

            self.commit_session()

            return f"id {anno_id} updated"
        except NoResultFound as e:
            return f"No result found for {anno_id}. Error: {e}"

    def delete_annotation(self, table_name: str,
                          annotation_ids: List[int]):
        """Delete annotations by ids

        Parameters
        ----------
        table_name : str
            name of table to delete from
        annotation_ids : List[int]
            list of ids to delete

        Returns
        -------

        Raises
        ------
        """
        AnnotationModel = self._cached_table(table_name)

        annotations = self.cached_session.query(AnnotationModel).filter(AnnotationModel.id.in_(annotation_ids)).all()
        if annotations:
            deleted_time = datetime.datetime.now()
            for annotation in annotations:
                annotation.deleted = deleted_time
                annotation.valid = False
            self.commit_session()
        else:
            return None
        return True
